export const input_group = "bk-input-group"
export const input = "bk-input"
export const spin_wrapper = "bk-spin-wrapper"
export const spin_btn_up = "bk-spin-btn-up"
export const spin_btn_down = "bk-spin-btn-down"
export default `:host(.slider-edit) .bk-input-group .bk-input{border:0;border-radius:0;min-height:0;padding-left:0;padding-right:0;font-weight:bold;box-shadow:unset;}:host(.slider-edit) .bk-input-group .bk-spin-wrapper{display:contents;}:host(.slider-edit) .bk-input-group .bk-spin-wrapper .bk-spin-btn-up{top:-6px;}:host(.slider-edit) .bk-input-group .bk-spin-wrapper .bk-spin-btn-down{bottom:3px;}`
